;nyquist plug-in
;version 3
;type analyze
;name "Album Track Gap Marker"
;action "Find track gaps..."
;author "Robert Haenggi"
;control marker "Mark Gap..." choice "Centroid,Center,Start,End,Region" 3
;
;; disabled for now
;;control additional "Additional Labels" choice "None,At Start,At Gap Start,At Gap End,At Gap Start & End" 0
;
;control threshold "Minimum Amplitude defining Silence" real "" -40 -70 -10
;control min-length "Minimum silent Time to detect" real "" 1.5 0.01 10
;control min-distance "Seconds to wait before new Labels are allowed" real "" 20.0 0.0 1800.0
(setf smoothing 4.0)
;;control smoothing "Smoothing Frequency (Hz)" real "" 4.0 0.01 20
(setf step (round (/ *sound-srate* 400)))
;;control step "Step size in samples" int "" 100 1 10000
(setf step 441)
;control text "Label text" string "" "Label"

(defun s-median (sig)
  (multichan-expand #'(lambda (v0)
     (let* ((lsr (snd-srate v0))
            (rsr (/ lsr)) 
            (dur (1- (snd-length v0  ny:all)))
            (first (sref v0 0))
            (last (snd-sref v0 (* rsr dur)))
            (v+1 (seq (snd-from-array 0 lsr (vector first))
                      (extract-abs 0 (* dur rsr) v0))) 
            (v-1 (seq (extract-abs rsr (* rsr (1+ dur)) v0)
                      (snd-from-array 0 lsr (vector last)))))
      (sim v-1 v0 v+1 (mult -1 (s-max  v+1 (s-max v-1 v0))) 
           (mult -1 (s-min v+1 (s-min v-1 v0)))))) sig))
;;

(defun total-gaps (env th)
  (let ((gaps 0) 
   (map (snd-samples
          (snd-inverse
            (integrate
              (s-abs ;(s-max (snd-const 0 0 (snd-srate env) (get-duration 1)) 
                (mult (/ (snd-srate s))
                  (slope
                    (snd-oneshot env th (- min-length (/ 1 (snd-srate env))))))))
           0 (snd-srate env))
         ny:all)))
   (dotimes (i (- (logand (length map) (lognot 1)) 2)  (1+ gaps))
      (if (> (- (aref map (+ i 2)) (aref map i)) min-distance) (incf gaps)))))
;;
(defun find-optimum (env)
  (do* ((db-step -16) 
        (border -80)
        (pivot (+ threshold db-step)  (+ pivot db-step))
        (gaps (total-gaps env threshold))
        (new-gaps (total-gaps env pivot)
                  (total-gaps env pivot)))
        ((or (<=  pivot -100) (and (= new-gaps gaps) (= border (1- pivot)))) 
         pivot)
   (cond
      ((/= new-gaps gaps)
          (setq border  pivot)
          (setq db-step (max 1 (abs (/ db-step 2)))))
      ((and (/= db-step -16) (= new-gaps gaps))
          (setq db-step (- (max 1 (abs (/ db-step 2)))))))
   (display "Try" gaps new-gaps pivot db-step border)))
;;;;
;;;; Main
;;;;
;; remove DC
(setf s (highpass2 s 50))
;; take highest value in all channels
(setf s 
    (if (arrayp s)
        (s-max (s-abs (aref s 0)) (s-abs (aref s 1)))
        (s-abs s)))
;; average the mono envelope
(setf s
  (mult 2
    (snd-avg s step step op-peak)))
;; filter out frequencies between ~4 Hz and sr/step Hz
;(setf s (lowpass8 s smoothing))
;; middle value out of three
(setf s (s-median s))
;; Force to detect at least a gap at the very beginning
;(setf s (sim  (snd-zero 0 (snd-srate s)) (at-abs (* 1 min-length) (cue s))))
;; convert to dB
(setf s (s-max (snd-const -128 0 (snd-srate s) (+ (get-duration 1) min-length)) (quantize (linear-to-db s) 1)))
;;
;;
(let* ((sr (snd-srate s))
       (minlen (* min-length sr))
       (troughs (sum  -128 (snd-avg (sum 128 s) (round  minlen)  1 op-peak)))       
       (labels ())
       (r-sr (recip sr))
       (start 0)
       (end 0)
       (valsum 0)
       (valprod 0)
       (centroid 0)
       (center 0)
       (nr 0)
       (latency (round (* sr (- 0 0.22))))
       (latency 0))
  (flet  ((transform (&rest args)
     (max 0(* r-sr (apply '+ (cons latency args))))))
     (print (snd-samples  (extract-abs 21 23.5 s) 2000))
     (do ((val (snd-fetch troughs ) (snd-fetch troughs))
          (min-val (1+ threshold))
          (count 0 (1+ count)))
         ((not val) (if (> (length labels) 0)
                        labels
                        "No silence found."))
       (cond
         ((> val threshold)
           (when (<= min-val threshold) 
               (setf min-val (1+ threshold))
               (case marker
                  (0; centroid
                     (push (list (transform centroid) label-text) labels))
                  (1 (setf  center (/ (+ start end) 2)); center
                     (push (list (transform  start center) label-text) labels))
                  (2; gap start 
                     (push (list (transform  start) label-text) labels))
                  (3; gap end 
                     (push (list (transform end) label-text) labels))
                  (4; gap as region label
                     (push (list (transform start)
                                 (transform  end) label-text) labels)))
               (if (and (> (length labels) 1) 
                        (< (- (caar labels) (caadr labels)) min-distance)) 
                   (pop labels))))
         ((<= val threshold)
           (cond
              ((< val min-val)
               (if (> val -60)
                   (setf start count))
               (psetq centroid (+ count (/ minlen 2))                      
                      label-text (format nil "~a" text)
                      end (+ (max centroid start) minlen)
                      min-val val))
              ((or (<= val -60) (<= val (+ min-val 3)))
               ; start will be the same, duration changes
               (setf end (+ count minlen)))
              (t nil)))))))